# CMake file
# Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2012
# Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2012
#
# See the file LICENSE.txt in the package base directory for details
#
# @file HelperMacros.cmake
#       Contains helpful macros.
#
# @author Tobias Hilbrich
# @date 16.03.2009

#=========================================================
# Macro from GTI
# Does:
#   Gets the name of a library, Gets the path to the 
#   P^nMPI patcher, Creates a CMake script in the binary
#   directory which executes the patcher for the given 
#   library. The result is placed in the P^nMPI library 
#   directory. The script is added as a install script.
#   (executed at install time)
#=========================================================
MACRO (GTI_MAC_PATCH_LIB targetname patcher)
    #TODO: test whether this works with Windows pathes (spaces and such)
    SET (lib ${CMAKE_SHARED_MODULE_PREFIX}${targetname}${CMAKE_SHARED_MODULE_SUFFIX})
    
    FILE (WRITE ${PROJECT_BINARY_DIR}/install-scripts/patch-${lib}.cmake 
        "EXECUTE_PROCESS (COMMAND mkdir -p ${CMAKE_INSTALL_PREFIX}/modules"
        "   RESULT_VARIABLE result"
        "   OUTPUT_VARIABLE output"
        "   ERROR_VARIABLE output)\n"
        "IF (result)\n"
        "   MESSAGE (FATAL_ERROR MKDIR MODULES \${output})\n"
        "ENDIF (result)\n"
        # return value of test is 0 or 1 whether the lib is newer then the patched one or not, also work with missing pached one
        "EXECUTE_PROCESS (COMMAND test ${LIBRARY_OUTPUT_PATH}/${lib} -ot ${CMAKE_INSTALL_PREFIX}/modules/${lib}"
        "   WORKING_DIRECTORY ${PROJECT_BINARY_DIR}/install-scripts"
        "   RESULT_VARIABLE tresult"
        "   OUTPUT_VARIABLE toutput"
        "   ERROR_VARIABLE toutput)\n"
        "IF (tresult)\n"
        "   MESSAGE (\"Patching ${lib}\")\n"
        "   EXECUTE_PROCESS (COMMAND ${patcher}"
        "       ${LIBRARY_OUTPUT_PATH}/${lib} ${CMAKE_INSTALL_PREFIX}/modules/${lib}"
        "       WORKING_DIRECTORY ${PROJECT_BINARY_DIR}/install-scripts"
        "       RESULT_VARIABLE result"
        "       OUTPUT_VARIABLE output"
        "       ERROR_VARIABLE output)\n"
        "   IF (result)\n"
        "       MESSAGE (FATAL_ERROR PATCHING \${output})\n"
        "   ENDIF (result)\n"
        "ELSE (tresult)\n"
        "   MESSAGE (\"patched lib Up-to-date ${lib}\")\n"
        "ENDIF (tresult)\n"
        )
    INSTALL (SCRIPT ${PROJECT_BINARY_DIR}/install-scripts/patch-${lib}.cmake)
ENDMACRO (GTI_MAC_PATCH_LIB)

#=========================================================
# Macro
# Does:
#   Takes a targetname and a list of source files.
#   Adds a module with given name and sources, installs 
#   patches and versions it.
# 
# language: in C, CXX, Fortran
#=========================================================
MACRO (GTI_MAC_ADD_MODULE targetname sources language)
    #Add target and its dependency on the patcher
    ADD_LIBRARY(${targetname} MODULE ${sources})
    
    #For Apple set that undefined symbols should be looked up dynamically
    #(On linux this is already the default)
    IF (APPLE)
        SET_TARGET_PROPERTIES(${targetname} PROPERTIES LINK_FLAGS "-undefined dynamic_lookup")
    ENDIF (APPLE)
    
    #Add avoid MPICXX header flags                                                                          
    IF ("${language}" STREQUAL "CXX")
        SET (TEMP "")
        FOREACH (skip_flag ${MPI_CXX_SKIP_FLAGS})
            SET (TEMP "${TEMP} ${skip_flag}")
        ENDFOREACH (skip_flag)
        SET_TARGET_PROPERTIES(${targetname} PROPERTIES COMPILE_FLAGS "${TEMP}")
    ENDIF ("${language}" STREQUAL "CXX")
    
    ##Not necessary: the script with the patcher is run anyways
    #Install it with reasonable file permissions
    #INSTALL(TARGETS ${targetname}
    #    PERMISSIONS 
    #        OWNER_READ OWNER_WRITE OWNER_EXECUTE 
    #        GROUP_EXECUTE GROUP_READ 
    #        WORLD_EXECUTE
    #    RUNTIME DESTINATION bin
    #    LIBRARY DESTINATION modules
    #    ARCHIVE DESTINATION modules
    #    )
        
    #Patch it during installation
    GTI_MAC_PATCH_LIB (
        ${targetname}
        ${PNMPI_PATCHER}
        )       
ENDMACRO (GTI_MAC_ADD_MODULE)


####################################################
## Macro featureTestMpi
##
## Creates a feature test with an MPI program. 
####################################################
MACRO (
    featureTestMpi 
        source # The source file name of the MPI program to test, must be in the folder cmakemodules/FeatureTests
        language # one of: C, CXX, Fortran
        successVar # Name of a variable to set to true iff the test was successful, will be set as a CACHE variable
        )
    
    #Organize the temporary src directory, copy source there
    SET (binDir "${CMAKE_CURRENT_BINARY_DIR}/${source}/BUILD")
    SET (srcDir "${CMAKE_CURRENT_BINARY_DIR}/${source}")
    CONFIGURE_FILE("${CMAKE_SOURCE_DIR}/cmakemodules/FeatureTests/${source}" "${srcDir}/${source}" COPYONLY)
    
    #Determine the compilers to use
    FOREACH (lan C CXX Fortran)
        SET (${lan}_compiler_to_use ${CMAKE_${lan}_COMPILER})
        IF (MPI_${lan}_COMPILER)
            SET (${lan}_compiler_to_use ${MPI_${lan}_COMPILER})
        ENDIF ()
    ENDFOREACH ()
    
    #Create CMakeLists.txt
    FILE(WRITE "${srcDir}/CMakeLists.txt" 
        "PROJECT (test ${language})\n"
        "cmake_minimum_required(VERSION 2.6)\n"
        "set(CMAKE_C_COMPILER \"${C_compiler_to_use}\")\n"
        "set(CMAKE_CXX_COMPILER \"${CXX_compiler_to_use}\")\n"
        "set(CMAKE_Fortran_COMPILER \"${Fortran_compiler_to_use}\")\n"
        "add_definitions(${skipPrepare})\n"
        "include_directories(\"${MPI_${language}_INCLUDE_PATH}\")\n"
        "add_executable(test \"${source}\")\n"
        "target_link_libraries(test ${MPI_${language}_LIBRARIES})\n"
        )
        
    #Try compile and preserve the result in a cached variable
    try_compile(${successVar} "${binDir}" "${srcDir}"
                     test
                     OUTPUT_VARIABLE output)
                     
    SET (${successVar} ${${successVar}} CACHE INTERNAL "")

    IF (GTI_VERBOSE)
        MESSAGE ("${output}")
    ENDIF (GTI_VERBOSE)
    
    #Add status
    SET (successStatus "success")
    IF (NOT ${successVar})
        SET (successStatus "failed")
    ENDIF (NOT ${successVar})
    
    MESSAGE (STATUS "Checking for ${source} ... ${successStatus}")

ENDMACRO (featureTestMpi)
