/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CProtMpiSplitModule.cpp generated from cprot_mpi_split_module.w
 *       Splits MPI processes into multiple sets of processes.
 *       Intention is to use one set(id:0) for the actual application
 *       and the remaining sets for tool processes (one set for
 *       each level of the tool).
 *       Further, a separate stack is used for each of the sets,
 *       to enable tools with distinct layouts. Also, the actual
 *       application calls are separated such that MPI_COMM_WORLD
 *       is replaced by a comm representing the application
 *       processes set.
 *       For all the tool process sets only an MPI_Init and an
 *       MPI_Finalize is called, afterwards an exit(0) is issued,
 *       so all tool startup should be done in MPI_Init and it should
 *       only return once a shutdown is desired.
 *
 * @author Tobias Hilbrich
 * @date 27.07.2009
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <vector>
#include <map>
#include <string>
#include <unistd.h>
#include <sstream>

#include <dlfcn.h>

#include <pnmpimod.h>
#include "GtiMacros.h"
#include "CProtMpiSplitModule.h"

/**
 * Global definitions and helpers.
 */
/*Name of split module @todo move to some common place, currently present twice*/
#define MUST_SPLIT_MODULE_NAME "split_processes"

struct process_set_t
{
	int size;
	int start_rank;
	int last_rank;
	MPI_Comm set_comm;
	int set_index;
	PNMPI_modHandle_t stack;
};

std::vector<process_set_t> g_sets;
std::map <std::pair<int, int> ,int> g_mappings; /**< Maps a (OwnSetIndex,CommID) pair to a SetIndex, usage is to match equal comm_ids to the right set indices.*/

#define MACRO_MPI_Comm(_c) \
{ if (_c == MPI_COMM_WORLD) _c = g_sets[0].set_comm; }

/**
 * Query functions.
 */
extern "C" int getMySetSize (int *size)
{
	int rank;
	PMPI_Comm_rank(MPI_COMM_WORLD,&rank);
	*size = -1;

	for (int i = 0;i < g_sets.size();i++)
	{
		if (rank >= g_sets[i].start_rank && rank <= g_sets[i].last_rank)
			*size = g_sets[i].size;
	}
	return PNMPI_SUCCESS;
}

extern "C" int getMySetComm (void *comm)
{
	int rank;
	PMPI_Comm_rank(MPI_COMM_WORLD,&rank);
	*((MPI_Comm*)comm) = MPI_COMM_NULL;

	for (int i = 0;i < g_sets.size();i++)
	{
		if (rank >= g_sets[i].start_rank && rank <= g_sets[i].last_rank)
			*((MPI_Comm*)comm) = g_sets[i].set_comm;
	}
	return PNMPI_SUCCESS;
}

extern "C" int getSetInfo (int commId, int* set_size, int* start_rank)
{
	int ownSetId;
	int setIdToUse;

	//get own set id
	int rank;
	PMPI_Comm_rank(MPI_COMM_WORLD,&rank);
	for (int i = 0;i < g_sets.size();i++)
	{
		if (rank >= g_sets[i].start_rank && rank <= g_sets[i].last_rank)
			ownSetId = i;
	}

	//find set id to use
	std::map<std::pair<int,int>, int >::iterator iter =
			g_mappings.find(std::make_pair(ownSetId, commId));

	if (iter != g_mappings.end())
		setIdToUse = iter->second; //New version use the mapping (if given)
	else
		setIdToUse = commId; //Old version gicen commId is the setId to use

	assert (setIdToUse < g_sets.size() && setIdToUse >= 0);

	*set_size = g_sets[setIdToUse].size;
	*start_rank = g_sets[setIdToUse].start_rank;

	return PNMPI_SUCCESS;
}

/**
 * PNMPI_ReistrationPoint
 */
extern "C" int PNMPI_RegistrationPoint()
{
  int err;

  /* register this module*/
  err=PNMPI_Service_RegisterModule(MUST_SPLIT_MODULE_NAME);
  assert (err == PNMPI_SUCCESS);

  return err;
}

#ifdef MUST_TIME
struct timeval gStart, gEnd;
#endif

extern "C" int MPI_Init(int * argc, char *** argv)
{
  int err;
  const char *inp;
  int split_mod;
  int num_sets;
  int world_size, rank;
  int curr_set_start = 0;
  PNMPI_Service_descriptor_t service;

#ifdef MUST_TIME
  gettimeofday(&gStart, NULL);
#endif

  /* Get self module */
  err = PMPI_Init(argc, argv);
  assert (err == MPI_SUCCESS);
  err = PMPI_Comm_size(MPI_COMM_WORLD, &world_size);
  assert (err == MPI_SUCCESS);
  err = PMPI_Comm_rank(MPI_COMM_WORLD, &rank);
  assert (err == MPI_SUCCESS);
  err = PNMPI_Service_GetModuleByName(MUST_SPLIT_MODULE_NAME,&split_mod);
  assert (err==PNMPI_SUCCESS);

  //enable a sleep to allow attaching with a debugger
  if (getenv("MUST_WAIT_AT_STARTUP") != NULL)
  {
	  std::cout << "Rank " << rank << " has pid " << getpid() << std::endl;
	  sleep (atoi(getenv("MUST_WAIT_AT_STARTUP")));
  }
  
  /*
   * EXPERIMENTAL, might be of use in the future. 
   * Idea is to automatically connect a gdbserver to some ranks to allow remote debugging
   * in unfriendly environments.
   */  
  if (getenv("MUST_START_GDBSERVER_BEGIN") != NULL && getenv("MUST_START_GDBSERVER_END") != NULL)
  {
  	  int begin = atoi(getenv("MUST_START_GDBSERVER_BEGIN"));
  	  int end = atoi(getenv("MUST_START_GDBSERVER_END"));
  
  	  if (rank >= begin && rank <= end)
  	  {  
	      char host[512];
	      gethostname(host, 512);
	      std::stringstream stream;
	      stream
	      	<< "gdbserver "
	      	<< host
	      	<< ":"
	      	<< rank + 10000
	      	<< " --attach "
	      	<< getpid ()
	      	<< " &"
	      	<< std::endl;
	      system (stream.str().c_str());
	      std::cout << "GDBSERVER for rank " << rank << " listens on " << host << ":" << rank + 10000 << std::endl;
	  }
  }

  /* Query for number of sets and their stacks */
  err=PNMPI_Service_GetArgument(split_mod,"num_sets",&inp);
  assert (err==PNMPI_SUCCESS);
  num_sets=atoi(inp);
  g_sets.resize(num_sets);

  for (int i = 0; i < num_sets; i++)
  {
	  char buf[512];
	  int  size;
	  PNMPI_modHandle_t stack;
	  int in_set = 0;

	  sprintf (buf,"size_%d",i);
	  err=PNMPI_Service_GetArgument(split_mod,buf,&inp);
	  assert (err==PNMPI_SUCCESS);
	  size=atoi(inp);

	  sprintf (buf,"stack_%d",i);
	  err=PNMPI_Service_GetArgument(split_mod,buf,&inp);
	  assert (err==PNMPI_SUCCESS);
	  err=PNMPI_Service_GetStackByName(inp,&stack);
	  assert (err==PNMPI_SUCCESS);

	  g_sets[i].size = size;
	  g_sets[i].start_rank = curr_set_start;
	  g_sets[i].last_rank = curr_set_start + size - 1;
	  g_sets[i].set_index = i;
	  g_sets[i].stack = stack;

	  if (rank >= g_sets[i].start_rank && rank <= g_sets[i].last_rank)
		  in_set = 1;

	  err = PMPI_Comm_split (MPI_COMM_WORLD, in_set, 0, &g_sets[i].set_comm);
	  assert (err == MPI_SUCCESS);

	  curr_set_start += size;
	  assert (world_size >= curr_set_start); //Sets must require no more processes than available!
  }

  /* Query for mappings */
  err=PNMPI_Service_GetArgument(split_mod,"num_mappings",&inp);
  if (err==PNMPI_SUCCESS)
  {
	  int numMappings = atoi(inp);

	  for (int i = 0; i < numMappings; i++)
	  {
		  char buf[512];
		  int  size;

		  sprintf (buf,"mapping%d",i);
		  err=PNMPI_Service_GetArgument(split_mod,buf,&inp);
		  assert (err==PNMPI_SUCCESS);

		  std::string mapping (inp);

		  int ownSetId, commId, setIdToUse;
		  std::string ownSetIdStr, commIdStr, setIdToUseStr;

		  //Own set id
		  size_t pos = mapping.find_first_of (':');
		  assert (pos != std::string::npos);
		  ownSetIdStr = mapping.substr (0, pos);

		  //comm id
		  size_t pos2 = mapping.find_first_of (':',pos+1);
		  assert (pos2 != std::string::npos);
		  commIdStr = mapping.substr (pos+1, pos2-pos-1);

		  //Set id to use
		  setIdToUseStr = mapping.substr (pos2+1, mapping.length()-pos2-1);

		  ownSetId = atoi (ownSetIdStr.c_str());
		  commId = atoi (commIdStr.c_str());
		  setIdToUse = atoi (setIdToUseStr.c_str());

		  g_mappings.insert (std::make_pair(std::make_pair(ownSetId, commId), setIdToUse));
	  }
  }

  /* Provide services to query for set information */
  sprintf(service.name,"SplitMod_getMySetSize");
  service.fct=(PNMPI_Service_Fct_t) getMySetSize;
  sprintf(service.sig,"p");
  err=PNMPI_Service_RegisterService(&service);
  assert (err == PNMPI_SUCCESS);

  sprintf(service.name,"SplitMod_getMySetComm");
  service.fct=(PNMPI_Service_Fct_t) getMySetComm;
  sprintf(service.sig,"p");
  err=PNMPI_Service_RegisterService(&service);
  assert (err == PNMPI_SUCCESS);

  sprintf(service.name,"SplitMod_getSetInfo");
  service.fct=(PNMPI_Service_Fct_t) getSetInfo;
  sprintf(service.sig,"ipp");
  err=PNMPI_Service_RegisterService(&service);
  assert (err == PNMPI_SUCCESS);

  /* Call MPI_Init for the correct stack */
  for (int i = 0; i < num_sets; i++)
  {
	  if (rank >= g_sets[i].start_rank && rank <= g_sets[i].last_rank)
	  {
		  //APPLICATION
		  if (i == 0)
		  {
			  err = PMPI_Init_NewStack(g_sets[i].stack,argc, argv);
			  assert (err == MPI_SUCCESS);
		  }
		  //Tool Processes
		  else
		  {
			  /*Init the MPI tool place, when it returns finalize and exit*/
			  err = PMPI_Init_NewStack(g_sets[i].stack,argc, argv);
			  assert (err == MPI_SUCCESS);

#ifdef MUST_TIME
			  gettimeofday(&gEnd, NULL);

			  std::cout << "Time Pre Init - Pre Finalize of Tool " << rank << " (usec): " << ((gEnd.tv_sec * 1000000 + gEnd.tv_usec) - (gStart.tv_sec * 1000000 + gStart.tv_usec)) << std::endl;
#endif

              /*
               * This does not calls the real MPI_Finalize,
               * it just passes the finalize to all modules of
               * the stack, we must still call the actual finalize
               * afterwards.
               */
			  err = PMPI_Finalize_NewStack(g_sets[i].stack);
			  assert (err == MPI_SUCCESS);

			  PMPI_Barrier(MPI_COMM_WORLD);

              /*
               * IMPORTANT:
               * ... prepare for ugliness!
               * Only an MPI_Fin from the application triggers the real MPI_Finalize!
               * We would need to get a non-patched PMPI_Finalize call here, while 
               * at the same time we need the patching for other symbols.
               * The current fix gets the "MPI_Finalize" symbol from the pnmpi_lib
               * which we assume to be there. Then we call the MPI_Fin in the pnmpi_lib,
               * this should emulate what happens when the app calls MPI_Fin.
               *
               * @todo for Windows this might be a bad idea, would we also get a copy
               *             of the already loaded module (good)? Or a new handle of a duplicated
               *             module (bad)?
               */
              void* theHandle = dlopen (PNMPI_LIB, RTLD_LAZY);
              int (*finP) ();
              finP = (int (*) ()) dlsym (theHandle, "MPI_Finalize");
	          err = finP ();
			  assert (err == MPI_SUCCESS);

			  exit(0);
		  }
	  }
  }

  return err;
}

extern "C" int MPI_Finalize()
{
  int err;

  int rank;
  PMPI_Comm_rank (MPI_COMM_WORLD,&rank);

  if (rank >= g_sets[0].start_rank && rank <= g_sets[0].last_rank)
  {
	  /*
       * This does not calls the real MPI_Finalize,
       * it just passes the finalize to all modules of
       * the stack, we must still call the actual finalize
       * afterwards.
       */
	  err = PMPI_Finalize_NewStack(g_sets[0].stack);
	  assert (err == MPI_SUCCESS);
	  PMPI_Barrier(MPI_COMM_WORLD);
  }

  /**
   * Now forward the MPI_Finalize, this works as
   * the application called an MPI_Finalize to bring
   * us here (not at in the case for the tool processes).
   */
  err=PMPI_Finalize();
  assert (err == MPI_SUCCESS);

  return err;
}

/*
 * -------------------------------------------
 */
/*GENERATED FUNCTIONS*/
/*
 * -------------------------------------------
 */
/* automatically generated wrapper code */
{{fnall fn_name MPI_Init MPI_Pcontrol MPI_Finalize}} {
  int _stack = g_sets[0].stack;
  {{apply_to_type MPI_Comm MACRO_MPI_Comm}}
  {{ret_val}} = P{{fn_name}}_NewStack({{list _stack {{args}} }});
}{{endfnall}}
