/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratSimpleDown.cpp
 *       Simple communication strategy.
 *
 * Sends all messages immediately and blocking.
 */

#include <assert.h>
#include <stdio.h>
#include <pnmpimod.h>
#include "CStratSimpleDown.h"
#include "GtiMacros.h"
#include <stdlib.h>

using namespace gti;

mGET_INSTANCE_FUNCTION(CStratSimpleDown)
mFREE_INSTANCE_FUNCTION(CStratSimpleDown)
mPNMPI_REGISTRATIONPOINT_FUNCTION(CStratSimpleDown)

const unsigned long CStratSimpleDown::myTokenShutdownSync = 0xFFFFFFFF;
const unsigned long CStratSimpleDown::myTokenMessage = 0xFFFFFFFE;
const unsigned long CStratSimpleDown::myTokenAck = 0xFFFFFFFD;

//=============================
// buf_free_function
//=============================
GTI_RETURN buf_free_function (void* free_data, unsigned long num_bytes, void* buf)
{
    delete [] (unsigned long*) buf;
    return GTI_SUCCESS;
}

//=============================
// CStratSimpleDown
//=============================
CStratSimpleDown::CStratSimpleDown (const char* instanceName)
: ModuleBase<CStratSimpleDown, CStratDownQueue> (instanceName)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //Either uses a regular comm protocol or a regular and a
    //fall-back one.
    assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    protocol = (I_CommProtocol*) subModInstances[0];
    if (subModInstances.size() == 2)
        fallback = (I_CommProtocol*) subModInstances[1];
    else
        fallback = NULL;

    //Set up internal data
    test_request = 0xFFFFFFFF;
    test_buff[0] = test_buff[1] = 0; /*Of no importance*/
}

//=============================
// ~CStratSimpleDown
//=============================
CStratSimpleDown::~CStratSimpleDown (void)
{
    /*Destroy sub modules*/
    if (protocol)
    	destroySubModuleInstance ((I_Module*)protocol);
    protocol = NULL;

    if (fallback)
        destroySubModuleInstance ((I_Module*)fallback);
    fallback = NULL;
}

//=============================
// shutdown
//=============================
GTI_RETURN CStratSimpleDown::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==Flush
    if (flush_behavior == GTI_FLUSH)
        flush ();

    // Remove protocol requests, to avoid sync msgs beeing catched by
    // outstanding requests !
    protocol->removeOutstandingRequests ();

    if (sync_behavior == GTI_SYNC)
    {
        unsigned long num,i;
        protocol->getNumChannels(&num);
        for (i = 0; i < num; i++)
        {
            unsigned long buf[2] = {myTokenShutdownSync,0};
            unsigned long length;

            //Send the ping
            protocol->ssend(buf,sizeof(unsigned long)*2,i);

            //receive the pong
            do
            {
                protocol->recv(buf,sizeof(unsigned long)*2,&length,i, NULL);
                assert (length == sizeof(unsigned long)*2);
                if (buf[0] != myTokenShutdownSync)
                {
                    //TODO give a warning here, received some other outstanding
                    //     message even though this strategy is being closed
                    //Receive the message, and discard it
                    std::cerr << "WARNING: In shutdown sync (StrategyUp) "
                              << "received an outstanding message!"
                              << std::endl;
                    unsigned long *temp_buf;
                    length = buf[1];
                    if (length % sizeof(unsigned long)) length = ((length/sizeof(unsigned long))+1)*sizeof(unsigned long);
                    temp_buf = new unsigned long[length/sizeof(unsigned long)];
                    protocol->recv(temp_buf,buf[1],&length,i, NULL);
                    delete [] temp_buf;
                }
            } while (buf[0] != myTokenShutdownSync);
        }
    }

    //Shutdown the protocols
    protocol->shutdown();
    if (fallback)
        fallback->shutdown();

    return GTI_SUCCESS;
}

//=============================
// broadcast
//=============================
GTI_RETURN CStratSimpleDown::broadcast (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	AddToQueue (buf, num_bytes, free_data, buf_free_function);
    	return GTI_SUCCESS;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==broadcast
    unsigned long num,i;
    protocol->getNumChannels(&num);
    for (i = 0; i < num; i++)
    {
        unsigned long temp_buf[2] = {myTokenMessage,
                                num_bytes};
        protocol->ssend(temp_buf,sizeof(unsigned long)*2,i);
        protocol->ssend(buf,num_bytes,i);
    }
    if (buf_free_function)
        (*buf_free_function) (free_data, num_bytes, buf);

    return GTI_SUCCESS;
}

//=============================
// getNumClients
//=============================
GTI_RETURN CStratSimpleDown::getNumClients (unsigned long *outNumClients)
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==getNumClients
    if (outNumClients)
        protocol->getNumChannels(outNumClients);

    return GTI_SUCCESS;
}

//=============================
// test
//=============================
GTI_RETURN CStratSimpleDown::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==test
    if (test_request == 0xFFFFFFFF)
    {
        protocol->irecv(test_buff, sizeof(unsigned long)*2, &test_request, RECV_ANY_CHANNEL);
    }

    int completed;
    unsigned long numbytes, channel;

    protocol->test_msg(test_request,&completed,&numbytes,&channel);

    if (completed)
    {
        assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message
        assert (test_buff[0] == myTokenMessage); //has to be a message

        unsigned long* buf;
        unsigned long  buf_size, recv_size;

#if 0
        if (test_buff[1] % sizeof(unsigned long) != 0)
            buf_size = (test_buff[1]/sizeof(unsigned long))+1;
        else
            buf_size = (test_buff[1]/sizeof(unsigned long));
#else
        buf_size = ((test_buff[1] - 1) / sizeof(unsigned long)) + 1;
#endif

        buf = new unsigned long[buf_size];

        protocol->recv(buf,test_buff[1],&recv_size,channel, NULL);

        assert (recv_size == test_buff[1]); //must be the size specified in first message

        //Save request as used up
        test_request = 0xFFFFFFFF;

        //save values
        if (out_flag)
        		*out_flag = 1;
        if (out_num_bytes)
        		*out_num_bytes = recv_size;
        if (out_buf)
        		*out_buf = buf;
        if (out_free_data)
        		*out_free_data = NULL;
        if (out_buf_free_function)
        		*out_buf_free_function = buf_free_function;
        if (outChannel)
        		*outChannel = channel;
    }
    else
    {
        //No message received
        *out_flag = 0;
    }

    return GTI_SUCCESS;
}

//=============================
// wait
//=============================
GTI_RETURN CStratSimpleDown::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{

    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==wait
    unsigned long numbytes, channel;

    if (test_request != 0xFFFFFFFF)
    {
        protocol->wait_msg(test_request,&numbytes,&channel);
    }
    else
    {
        protocol->recv(test_buff,sizeof(unsigned long)*2,&numbytes,RECV_ANY_CHANNEL,&channel);
    }

    assert (numbytes == sizeof(unsigned long)*2); //has to be a token,length message
    assert (test_buff[0] == myTokenMessage); //has to be a message

    unsigned long* buf;
    unsigned long  buf_size, recv_size;

#if 0
    if (test_buff[1] % sizeof(unsigned long) != 0)
        buf_size = (test_buff[1]/sizeof(unsigned long))+1;
    else
        buf_size = (test_buff[1]/sizeof(unsigned long));
#else
    buf_size = ((test_buff[1] - 1) / sizeof(unsigned long)) + 1;
#endif

    buf = new unsigned long[buf_size];

    protocol->recv(buf,test_buff[1],&recv_size,channel, NULL);

    assert (recv_size == test_buff[1]); //must be the size specified in first message

    //Save request as used up
    test_request = 0xFFFFFFFF;

    //save values
    if (out_num_bytes)
    		*out_num_bytes =	 recv_size;
    if (out_buf)
    		*out_buf = buf;
    if (out_free_data)
    		*out_free_data = NULL;
    if (out_buf_free_function)
    		*out_buf_free_function = buf_free_function;
    if (outChannel)
    		*outChannel = channel;
    return GTI_SUCCESS;
}

//=============================
// acknowledge
//=============================
GTI_RETURN CStratSimpleDown::acknowledge (
		unsigned long channel
        )
{
	if (!protocol->isConnected())
	{
	    	return GTI_ERROR_NOT_INITIALIZED;
	}

	unsigned long temp_buf[2] = {myTokenAck,0};
	protocol->ssend(temp_buf,sizeof(unsigned long)*2,channel);
	return GTI_SUCCESS;
}

//=============================
// flush
//=============================
GTI_RETURN CStratSimpleDown::flush (
        void
        )
{

    //Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //flush
    //No action needed
    //->broadcast uses ssend

    return GTI_SUCCESS;
}

//=============================
// enforce_blocking_comm
//=============================
GTI_RETURN CStratSimpleDown::enforce_blocking_comm (
        void
        )
{

    //Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //enforce_blocking_comm
    //No action needed
    //->broadcast uses ssend

    return GTI_SUCCESS;
}

//=============================
// activate_fallback_comm
//=============================
GTI_RETURN CStratSimpleDown::activate_fallback_comm (
        void
        )
{

    //not supported
    assert (0);

    return GTI_SUCCESS;
}

/*EOF*/
