/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file CStratThreadedDown.cpp
 *        Application phase aware communication strategy.
 *
 * Provides the interface functionality with an implementation
 * that tries to avoid any comm. during phases in which the
 * application is communicating.
 *
 * @author Tobias Hilbrich
 * @date 24.08.2009
 */

#include <assert.h>
#include <stdio.h>
#include <mpi.h>
#include <pnmpimod.h>
#include "CStratThreadedDown.h"
#include "GtiMacros.h"
#include <stdlib.h>

using namespace gti;

mGET_INSTANCE_FUNCTION(CStratThreadedDown)
mFREE_INSTANCE_FUNCTION(CStratThreadedDown)
mPNMPI_REGISTRATIONPOINT_FUNCTION(CStratThreadedDown)

//=============================
// CStratThreadedDown
//=============================
CStratThreadedDown::CStratThreadedDown (const char* instanceName)
: ModuleBase<CStratThreadedDown, CStratDownQueue> (instanceName),
  CStratAggregateReceiver (&protocol, BUF_LENGTH)
{
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //Either uses a regular comm protocol or a regular and a
    //fall-back one.
    assert ((subModInstances.size() == 1) || (subModInstances.size() == 2));

    protocol = (I_CommProtocol*) subModInstances[0];
    if (subModInstances.size() == 2)
        fallback = (I_CommProtocol*) subModInstances[1];
    else
        fallback = NULL;

    //Set up internal data

}

//=============================
// ~CStratThreadedDown
//=============================
CStratThreadedDown::~CStratThreadedDown (void)
{
    /*Destroy sub modules*/
    if (protocol)
        destroySubModuleInstance ((I_Module*)protocol);

    if (fallback)
        destroySubModuleInstance ((I_Module*)fallback);

    //Free internal data
}


//=============================
// shutdown
//=============================
GTI_RETURN CStratThreadedDown::shutdown (
        GTI_FLUSH_TYPE flush_behavior,
        GTI_SYNC_TYPE sync_behavior
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
    	ProcessQueue ();

    //==shutdown
    if (flush_behavior == GTI_FLUSH)
        flush ();

    // Remove protocol requests, to avoid sync msgs beeing catched by
    // outstanding requests !
    protocol->removeOutstandingRequests ();

    if (sync_behavior == GTI_SYNC)
    {
        unsigned long num,i;
        protocol->getNumChannels(&num);
        for (i = 0; i < num; i++)
        {
            unsigned long buf[2] = {myTokenShutdownSync,0};
            unsigned long length;

            //Send the ping
            protocol->ssend(buf,sizeof(unsigned long)*2,i);

            //receive the pong
            do
            {
                CStratAggregateInfo* aggregate = get_free_aggregate();
                protocol->recv(aggregate->buf, BUF_LENGTH, &length, i, NULL);
                assert (length >= sizeof(unsigned long)*2);
                if (((unsigned long*)aggregate->buf)[0] != myTokenShutdownSync)
                {
                    //TODO give a warning here, received some other outstanding
                    //     message even though this strategy is being closed
                    //Receive the message, and discard it
                    std::cerr << "WARNING: In shutdown sync (StrategDown) "
                            << "received an outstanding message!"
                            << std::endl;

                    if (((unsigned long*)aggregate->buf)[0] != myTokenLongMsg)
                    {
                        unsigned long *temp_buf;
                        length = ((unsigned long*)aggregate->buf)[1];
                        if (length % sizeof(unsigned long)) length = ((length/sizeof(unsigned long))+1)*sizeof(unsigned long);
                        temp_buf = new unsigned long[length/sizeof(unsigned long)];
                        protocol->recv(temp_buf, ((unsigned long*)aggregate->buf)[1], &length, i, NULL);
                        delete [] temp_buf;
                    }

                    myFreeAggregates.push_back(aggregate);
                }
            } while (buf[0] != myTokenShutdownSync);
        }
    }

    //Shutdown the protocols
    protocol->shutdown();
    if (fallback)
        fallback->shutdown();
    return GTI_SUCCESS;
}

//=============================
// broadcast
//=============================
GTI_RETURN CStratThreadedDown::broadcast (
        void* buf,
        unsigned long num_bytes,
        void* free_data,
        GTI_RETURN (*buf_free_function) (void* free_data, unsigned long num_bytes, void* buf)
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        AddToQueue (buf, num_bytes, free_data, buf_free_function);
        return GTI_SUCCESS;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==broadcast
    unsigned long num,i;
    protocol->getNumChannels(&num);
    for (i = 0; i < num; i++)
    {
        unsigned long temp_buf[2] = {myTokenMessage,
                num_bytes};
        protocol->ssend(temp_buf,sizeof(unsigned long)*2,i);
        protocol->ssend(buf,num_bytes,i);
    }
    if (buf_free_function)
        (*buf_free_function) (free_data, num_bytes, buf);
    return GTI_SUCCESS;
}

//=============================
// getNumClients
//=============================
GTI_RETURN CStratThreadedDown::getNumClients (unsigned long *outNumClients)
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==getNumClients
    if (outNumClients)
        protocol->getNumChannels(outNumClients);

    return GTI_SUCCESS;
}

//=============================
// test
//=============================
GTI_RETURN CStratThreadedDown::test (
        int* out_flag,
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
    	return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==test

    //**
    //** (1) we still have an open aggregate with remaining messages
    //**
    if (myOpenAggregate)
    {
        GTI_RETURN ret = msg_from_open_aggregate (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
        return ret;
    }

    //**
    //** (2) Do we already have a request for test calls ?
    //**
    if (myTestRequest == 0xFFFFFFFF)
    {
    		myTestAggregate = get_free_aggregate();
        protocol->irecv(myTestAggregate->buf, BUF_LENGTH, &myTestRequest, RECV_ANY_CHANNEL);
    }

    //**
    //** (3) Test for completion of the request
    //**
    int completed;
    unsigned long numbytes, channel;

    protocol->test_msg(myTestRequest,&completed,&numbytes,&channel);

    //**
    //** (4) Handle new aggregate if the test succeeded
    //**
    if (completed)
    {
        assert (numbytes >= sizeof(unsigned long)*2); //each message has at least two entries
        assert (((unsigned long*)myTestAggregate->buf)[0] == myTokenMessage ||
        		((unsigned long*)myTestAggregate->buf)[0] == myTokenLongMsg    ); //has to be a message

        //Save request as used up
        myTestRequest = 0xFFFFFFFF;

        if (((unsigned long*)myTestAggregate->buf)[0] == myTokenMessage)
        {
				//Received an aggregate
				myOpenAggregate = myTestAggregate;
				myOpenAggregate->current_position = 2*sizeof(unsigned long);
				myOpenAggregate->num_msgs_left = ((unsigned long*)myOpenAggregate->buf)[1];
				myOpenAggregate->num_in_use = 0;
				myOpenAggregate->channel = channel;
				myOpenAggregate->instance = this;
				myTestAggregate = NULL;

				//return a message from the new aggregate
				GTI_RETURN ret = msg_from_open_aggregate (out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
				return ret;
        }
        else
        {
				//Received info for a long message
				GTI_RETURN ret = long_msg_from_info (
						((unsigned long*)myTestAggregate->buf)[1], channel,
						out_flag, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);

				myFreeAggregates.push_back (myTestAggregate);
				myTestAggregate = NULL;
				return ret;
        }
    }
    else
    {
        //No message received
        if (out_flag)
        	*out_flag = 0;
    }

    return GTI_SUCCESS;
}

//=============================
// wait
//=============================
GTI_RETURN CStratThreadedDown::wait (
        unsigned long* out_num_bytes,
        void** out_buf,
        void** out_free_data,
        GTI_RETURN (**out_buf_free_function) (void* free_data, unsigned long num_bytes, void* buf),
        unsigned long *outChannel
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==wait

    //**
    //** (1) we still have an open aggregate with remaining messages ?
    //**
    if (myOpenAggregate)
    {
        GTI_RETURN ret = msg_from_open_aggregate (NULL, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
        return ret;
    }

    //**
    //** (2) Wait for a message
    //**
    unsigned long numbytes, channel;

    if (myTestRequest != 0xFFFFFFFF)
    {
        protocol->wait_msg(myTestRequest,&numbytes,&channel);
    }
    else
    {
    		myTestAggregate = get_free_aggregate();
        protocol->recv(myTestAggregate->buf, BUF_LENGTH, &numbytes, RECV_ANY_CHANNEL, &channel);

        //DEBUG
        //std::cout << "RECVD NEW AGGREGATE: [" << numbytes << "]" << std::endl;
        //ENDDEBUG
    }

    //**
    //** (3) Handle new aggregate
    //**
    assert (numbytes >= sizeof(unsigned long)*2); //each message has at least two entries
    assert (((unsigned long*)myTestAggregate->buf)[0] == myTokenMessage ||
    		((unsigned long*)myTestAggregate->buf)[0] == myTokenLongMsg    ); //has to be a message

    //Save request as used up
    myTestRequest = 0xFFFFFFFF;

    GTI_RETURN ret;
    if (((unsigned long*)myTestAggregate->buf)[0] == myTokenMessage)
    {
        //Received an aggregate
        myOpenAggregate = myTestAggregate;
        myOpenAggregate->current_position = 2*sizeof(unsigned long);
        myOpenAggregate->num_msgs_left = ((unsigned long*)myOpenAggregate->buf)[1];
        myOpenAggregate->num_in_use = 0;
        myOpenAggregate->channel = channel;
        myOpenAggregate->instance = this;
        myTestAggregate = NULL;

        //return a message from the new aggregate
        ret = msg_from_open_aggregate (NULL, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);
    }
    else
    {
        //Received info for a long message
        ret = long_msg_from_info (
                ((unsigned long*)myTestAggregate->buf)[1], channel,
                NULL, out_num_bytes, out_buf, out_free_data, out_buf_free_function, outChannel);

        myFreeAggregates.push_back (myTestAggregate);
        myTestAggregate = NULL;
    }

    return ret;
}

//=============================
// acknowledge
//=============================
GTI_RETURN CStratThreadedDown::acknowledge (
		unsigned long channel
        )
{
	/*Nothing to do*/
	return GTI_SUCCESS;
}

//=============================
// flush
//=============================
GTI_RETURN CStratThreadedDown::flush (
        void
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
      	ProcessQueue ();

    //==flush
    //No action needed
    //->broadcast uses ssend
    return GTI_SUCCESS;
}

//=============================
// enforce_blocking_comm
//=============================
GTI_RETURN CStratThreadedDown::enforce_blocking_comm (
        void
        )
{
    //==Check whether a connection exists
    if (!protocol->isConnected())
    {
        return GTI_ERROR_NOT_INITIALIZED;
    }

    //==Check whether there are queued sends, process if so
    if (hasQueueEntries())
        ProcessQueue ();

    //==enforce_blocking_comm
    //No action needed
    //->broadcast uses ssend

    return GTI_SUCCESS;
}

//=============================
// activate_fallback_comm
//=============================
GTI_RETURN CStratThreadedDown::activate_fallback_comm (
        void
        )
{
    //not supported
    assert (0);

    return GTI_SUCCESS;
}

/*EOF*/
