/*
 * GTI (Generic Tool Infrastructure)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2009-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2009-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file ThreadedMpiPlace.cpp
 *       A placement driver that is started as an MPI
 *       application and uses multi-threading to allow
 *       concurrent communication (loggs, mgmt, trace).
 *
 *
 * @author Tobias Hilbrich
 */

#include <assert.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <sys/time.h>

#include <pnmpimod.h>
#include "ThreadedMpiPlace.h"
#include "GtiMacros.h"
#include "SuspensionBufferTree.h"

#ifdef GTI_VT
#include <vt_user.h>
#include <dlfcn.h>

void* theHandle = 0;

typedef void (*startP) (const char* name, const char* file, int lno);
typedef void (*endP) (const char* name);

void VT_User_start__(const char* name, const char* file, int lno)
{
	if (!theHandle)
		theHandle = dlopen ("libvt-mpi.so", RTLD_LAZY);

	startP fp = (startP) dlsym (theHandle, "VT_User_start__");
	fp (name, file, lno);
}

void VT_User_end__(const char* name)
{
	endP fp = (endP) dlsym (theHandle, "VT_User_end__");
	fp (name);
}
#endif /*GTI_VT*/


using namespace gti;

mGET_INSTANCE_FUNCTION (ThreadedMPIPlace)
mFREE_INSTANCE_FUNCTION (ThreadedMPIPlace)
mPNMPI_REGISTRATIONPOINT_FUNCTION(ThreadedMPIPlace)

//=============================
// getInstanceName
//=============================
GTI_RETURN getInstanceName (const char **instanceName)
{
	//get self module, get name of own instance
	int err = PNMPI_SUCCESS;
	PNMPI_modHandle_t modHandle;

	err=PNMPI_Service_GetModuleSelf(&modHandle);
	assert (err == PNMPI_SUCCESS);

	/* get own module name */
	char temp[64];
	sprintf (temp, "instanceToUse");
	err=PNMPI_Service_GetArgument(modHandle, temp, instanceName);

	if (err != PNMPI_SUCCESS)
	{
		std::cerr << "Error: threaded mpi place module needs a PnMPI argument named \"instanceToUse\" that lists a valid instance name to be used as instance." << std::endl;
		return GTI_ERROR;
	}

	return GTI_SUCCESS;
}

//=============================
// MPI_Init Wrapper
//=============================
extern "C" int MPI_Init (int* argc, char*** argv)
{
	//init MPI first
    int err = PMPI_Init (argc, argv);

    //get name of instance to use
    const char *instanceName;
    GTI_RETURN ret = getInstanceName (&instanceName);
    assert (ret == GTI_SUCCESS);

    //Initialize and run place
    ThreadedMPIPlace* place;
    place = ThreadedMPIPlace::getInstance(instanceName);
    assert(place);
    place->run();

    return err;
}

//=============================
// MPI_Finalize Wrapper
//=============================
extern "C" int MPI_Finalize (void)
{
	//get name of instance to use
	const char *instanceName;
	GTI_RETURN ret =getInstanceName (&instanceName);
	assert (ret == GTI_SUCCESS);

    ThreadedMPIPlace* place;
    place = ThreadedMPIPlace::getInstance(instanceName);
    ThreadedMPIPlace::freeInstanceForced (place);

    return PMPI_Finalize ();
}

//=============================
// ThreadedMPIPlace
//=============================
ThreadedMPIPlace::ThreadedMPIPlace (const char* instanceName)
    : ModuleBase<ThreadedMPIPlace, I_Module> (instanceName),
      myWrapper (NULL),
      myIntraRecv (NULL)
{
    //add the id to all sub modules
    int rank;
    PMPI_Comm_rank (MPI_COMM_WORLD, &rank);
    char temp[64];
    sprintf (temp,"%d",rank);
    addDataToSubmodules ("id", temp);

    //create sub modules
    std::vector<I_Module*> subModInstances;
    subModInstances = createSubModuleInstances ();

    //save sub modules
    assert (subModInstances.size() >= 2);
    myTraceRecv = (I_CommStrategyDown*) subModInstances[0];
    myReceival = (I_PlaceReceival*) subModInstances[1];

    //Lets determine the number of  additional modules we have
    //a) Module for intra communication
    int intraModuleIndex = -1; //Index into module array that is a intra comm strategy -1 for no intra comm strategy
    std::map<std::string, std::string> data = getData();
    if (data.find("intra_strat_index") != data.end ())
        intraModuleIndex = atoi(data.find("intra_strat_index")->second.c_str());

    if (intraModuleIndex >= 0)
        myIntraRecv = (I_CommStrategyIntra*) subModInstances[intraModuleIndex];

    //b) Module for wrapper
    int wrapperModuleIndex = 2;
    if (intraModuleIndex == 2)
        wrapperModuleIndex++;

    if (subModInstances.size() > wrapperModuleIndex)
        myWrapper = (I_Module*) subModInstances[wrapperModuleIndex];

}

//=============================
// ~ThreadedMPIPlace
//=============================
ThreadedMPIPlace::~ThreadedMPIPlace (void)
{
    if (myTraceRecv)
    {
        myTraceRecv->shutdown(GTI_FLUSH, GTI_SYNC);
        destroySubModuleInstance ((I_Module*) myTraceRecv);
        myTraceRecv = NULL;
    }

    if (myReceival)
        destroySubModuleInstance ((I_Module*) myReceival);
    myReceival = NULL;

    if (myWrapper)
        destroySubModuleInstance ((I_Module*) myWrapper);
    myWrapper = NULL;

    if (myIntraRecv)
    {
        myIntraRecv->shutdown(GTI_FLUSH);
        destroySubModuleInstance ((I_Module*) myIntraRecv);
        myIntraRecv = NULL;
    }
}

//=============================
// run
//=============================
void ThreadedMPIPlace::run (void)
{
#ifdef GTI_VT
	VT_TRACER("ThreadedMPIPlace::run");
#endif /*GTI_VT*/

	GTI_RETURN (*free_function) (void* free_data, unsigned long num_bytes, void* buf);
	int flag;
	unsigned long   size;
	void            *buf, *free_data;
	unsigned long   numClients, numChannels;
	unsigned long   numOps = 0;
	GTI_RETURN     ret;
	unsigned long fromChannel;
	SuspensionBufferTree *tree = NULL,
			*node,
			*firstSuspendedNode = NULL,
			*firstNonEmptyQ = NULL;
	unsigned long currTimeout = 0;
	const unsigned long timeoutThreshold = 5000000; //5s

#ifdef GTI_DEBUG
	//Debug
	int gRank;
	MPI_Comm_rank (MPI_COMM_WORLD, &gRank);
	char temp[256];
	sprintf (temp, "place_channel_tree_%d.dot", gRank);
	std::ofstream out (temp);
#endif

	//get number of clients
	myTraceRecv->getNumClients(&numClients);
	numChannels = numClients;
#ifdef GTI_DEBUG
	printf ("\n\nCLIENTS: %ld\n\n",numClients);
#endif

    //Loop until all channels signaled that they finished
	while (numClients)
	{
		bool receiveRecord = true; //Used to determine whehter it is necessary to receive a new record or not
		bool hasRecordToProcess = false;
		bool hadIntraEvent = false;
		RecordInfo processRecord;

		// 1) Check whether we have any records in our queues that could be processed
		//     If so select such a record as record to process
		//================================================
		if (tree)
		{
			if (tree->getQueuedRecord (&processRecord))
			{
#ifdef GTI_VT
			VT_TRACER("ThreadedMPIPlace::run::gotQueuedRecord");
#endif /*GTI_VT*/
				//Set that there is a record to process ... (we need not receive something)
				receiveRecord = false;
				hasRecordToProcess = true;
			}
		}

		// 2) No queued records left, first try intra communication (if available)
		//================================================
		if (receiveRecord && myIntraRecv)
		{
		    if (!receiveAndProcessIntraLayerEvent (&hadIntraEvent))
		        break; //If it fails we abort
		    //IMPORTANT: we do not do a long wait for initialization and we do not influence timeout here!
		    //                    I think this is good, but didn't give this too much thought yet.
		}

		// 3) No queued records left, receive a new record (inter layer communication)
		//================================================
		if (receiveRecord)
		{
#ifdef GTI_VT
			{
			VT_TRACER("ThreadedMPIPlace::run::test");
#endif /*GTI_VT*/

			ret = myTraceRecv->test(&flag, &size, &buf, &free_data, &free_function, &fromChannel);
			//ret = myTraceRecv->wait(&size, &buf, &free_data, &free_function, &fromChannel);
#ifdef GTI_VT
			}
#endif /*GTI_VT*/

			if (ret == GTI_ERROR)
			{
				std::cerr << "Place: error while receiving, communication strategy returned error." << std::endl;
				break;
			}

			if (ret == GTI_SUCCESS) //Necessary as strategy may also return it is not ready yet ...
			{
			    if (flag)
			    {
#ifdef GTI_VT
			        VT_TRACER("ThreadedMPIPlace::run::evaluateNewRecord");
#endif /*GTI_VT*/
			        I_ChannelId *channelId;
			        bool isFinalizer;

			        //Get and update channel id for this record
			        if (myReceival->getUpdatedChannelId(buf, size, fromChannel, numChannels, &channelId, &isFinalizer) != GTI_SUCCESS)
			            break; //Abort main loop in case of some error

			        //Is it the last finalizer?
			        if (isFinalizer && numClients == 1)
			        {
			            //We need to fininsh up intra layer communication before we continue
			            if (!finishIntraCommunication ())
			                break;
			        }

			        //Store information on the record
			        processRecord.buf = buf;
			        processRecord.free_data = free_data;
			        processRecord.num_bytes = size;
			        processRecord.recordChannelId = channelId;
			        processRecord.buf_free_function = free_function;

			        //Allocate tree if this is the first channel id we got
			        if (!tree) tree = new SuspensionBufferTree (channelId->getNumUsedSubIds()-1, channelId->getSubIdNumChannels(channelId->getNumUsedSubIds()-1));

			        //==If we have any suspension, we must check whether it is allowed to process the record
			        if (!tree->hasAnySuspension())
			        {
			            hasRecordToProcess = true;
			        }
			        else
			        {
			            //Find the node for this channel id in the channel id tree
			            firstSuspendedNode = firstNonEmptyQ = NULL;
			            node = tree->getNode(channelId, &firstSuspendedNode, &firstNonEmptyQ);

			            //Is the channel of the received record ok for processing ?
			            //If there is any non-empty queue, then no
			            if (firstNonEmptyQ)
			            {
			                //If there was a non empty queue on the way, add to that node, otherwise add to the leaf node
			                firstNonEmptyQ->pushBack(processRecord);

#ifdef GTI_DEBUG
			                ////Debug:
			                //std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
			                out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Added to queue from: "<< channelId->toString() << "\"];" << std::endl << "}" << std::endl;
			                tree->printAsDot(out);
#endif
			            }
			            else
			                if (node->hasAnySuspension() || firstSuspendedNode)
			                {
			                    //Can't process at the moment, has suspended sub-nodes or parents!
			                    node->pushBack(processRecord);

#ifdef GTI_DEBUG
			                    ////Debug:
			                    //std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
			                    out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Added to queue from(descendant blocked): "<< channelId->toString() << "\"];" << std::endl << "}" << std::endl;
			                    tree->printAsDot(out);
#endif
			                }
			                else
			                    if (node->getQueueSize() || node->getChildQueueSize())
			                    {
			                        //Can't process at the moment, there is a queued event in the childs that needs to be processed first
			                        node->pushBack(processRecord);

#ifdef GTI_DEBUG
			                        ////Debug:
			                        //std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
			                        out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Added to queue from(descendant has queue entry): "<< channelId->toString() << "\"];" << std::endl << "}" << std::endl;
			                        tree->printAsDot(out);
#endif
			                    }
			                    else
			                    {
			                        hasRecordToProcess = true;
			                    }
			        }//If (!tree->hasAnySuspension())
			    }//test had flag=true
			    else //flag
			    {
			        //we did not receive something ...
			        if (!hadIntraEvent)
			        {
#ifdef GTI_VT
			            VT_TRACER("ThreadedMPIPlace::run::sleep");
#endif /*GTI_VT*/

			            struct timeval gStart, gEnd;
			            gettimeofday(&gStart, NULL);
			            usleep (10);
			            gettimeofday(&gEnd, NULL);
			            currTimeout += ((gEnd.tv_sec * 1000000 + gEnd.tv_usec) - (gStart.tv_sec * 1000000 + gStart.tv_usec));

			            //Timeout threshold reached ?
			            if (currTimeout > timeoutThreshold)
			            {
			                currTimeout = 0;

			                //							if (tree && tree->hasAnySuspension())
			                //							{
			                myReceival->timeOutReductions();
			                if (tree)
			                {
			                    //Send missing acknowledge signals!
			                    std::list<long> list = tree->getChildsIndicesWithSuspension ();
			                    std::list<long>::iterator chanIter;
			                    for (chanIter = list.begin(); chanIter != list.end(); chanIter++)
			                    {
			                        /*See note in the other acknowledges*/
			                        myTraceRecv->acknowledge (*chanIter);
			                    }

			                    //Clear suspensions
			                    tree->removeAllSuspensions();
			                }

#ifdef GTI_DEBUG
			                //DEBUG:
			                std::cout << "Timed out!" << std::endl;
			                out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Time out\"];" << std::endl << "}" << std::endl;
			                if (tree) tree->printAsDot(out);
#endif
			                //							}
			            }
			        }//We did not had an inter event and no intra event too -> sleep
			    }//if flag
			}//if GTI_SUCCESS
			else
			{
			    //Input strategy not ready yet, wait a bit ...
			    usleep(500);
			}//if-else-if for return of wait call
		}//If need to receive record

		//4) Process a record if there was a valid record for processing
		//================================================
		if (hasRecordToProcess)
		{
#ifdef GTI_VT
			VT_TRACER("ThreadedMPIPlace::run::process");
#endif /*GTI_VT*/
			//reset timeout
			currTimeout = 0;

			//process
			bool wasChannelClosed = false;
			std::list<I_ChannelId*> reopenedChannels;
			std::list<I_ChannelId*>::iterator iter;
			int fromChannel = processRecord.recordChannelId->getSubId(processRecord.recordChannelId->getNumUsedSubIds()-1);

			if (myReceival->ReceiveRecord(
					processRecord.buf,
					processRecord.num_bytes,
					processRecord.free_data,
					processRecord.buf_free_function,
					&numClients,
					processRecord.recordChannelId,
					&wasChannelClosed,
					&reopenedChannels) != GTI_SUCCESS)
				break; //Abort main loop in case of some error

			//evaluate suspensions
			if (wasChannelClosed)
			{
				node = tree->getNode(processRecord.recordChannelId, NULL, NULL);
				node->setSuspension(true);
			}
			else
			{
			    //Done, acknowledge now!
			    /*
			     *Note: that we only achnowledge if an event is not involved in a reduction
			     *         or its reduction succeeded has concsequences. With this layout we
			     *         can only perform a reduction if we only need one event from each
			     *         channel. I.e. we are able to reduce some event on each level!
			     *         Usually that shouldn't be a problem, but it can cause unexpected behavior.
			     */
			    myTraceRecv->acknowledge (fromChannel);
			}

			for (iter = reopenedChannels.begin(); iter != reopenedChannels.end(); iter++)
			{
				I_ChannelId *id = *iter;
				if (!id) continue;

				node = tree->getNode(id, NULL, NULL);
				node->setSuspension(false);

				//Send an acknowledge of a finally processed event
				/*See note in the other acknowledge*/
				myTraceRecv->acknowledge (id->getSubId(id->getNumUsedSubIds()-1));

				delete (id);
			}

#ifdef GTI_DEBUG
			////Debug:
			//std::cout << "Test: " << processRecord.recordChannelId->toString() << std::endl;
			out << "digraph temp" << std::endl << "{" << std::endl << "myNode [label=\"Processed from: "<< processRecord.recordChannelId->toString() << "\"];" << std::endl << "}" << std::endl;
			tree->printAsDot(out);
#endif

			delete (processRecord.recordChannelId);
			numOps++;
		}//if has record to process
	}//while channels left

#ifdef GTI_DEBUG
	//Debug
	std::cout << "NumOps: " << numOps << std::endl;
#endif

	//Delete tree
	if (tree) delete (tree);

}//run method

//=============================
// receiveAndProcessIntraLayerEvent
//=============================
bool ThreadedMPIPlace::receiveAndProcessIntraLayerEvent (bool *hadEvent)
{
    GTI_RETURN (*free_function) (void* free_data, unsigned long num_bytes, void* buf);
    int flag;
    unsigned long   size;
    void            *buf, *free_data;
    GTI_RETURN     ret;
    unsigned long fromChannel;

    if (hadEvent)
        *hadEvent = false;

    if (!myIntraRecv)
        return true;

    ret = myIntraRecv->test(&flag, &fromChannel, &size, &buf, &free_data, &free_function);

    if (ret == GTI_ERROR)
    {
        std::cerr << "Place: error while receiving, communication strategy returned error." << std::endl;
        return false;
    }

    if (ret == GTI_SUCCESS && flag)
    {
        if (myReceival->ReceiveIntraRecord(
                buf,
                size,
                free_data,
                free_function,
                fromChannel) != GTI_SUCCESS)
            return false; //Abort main loop in case of some error

        if (hadEvent)
            *hadEvent = true;
    }

    return true;
}

//=============================
// finishIntraCommunication
//=============================
bool ThreadedMPIPlace::finishIntraCommunication (void)
{
    if (!myIntraRecv)
        return true;

    bool hadEvent = false;
    bool isFinished = false;

    //Loop till intra communication is finished
    do
    {
        //Receive as much as we can right now
        do
        {
            if (!receiveAndProcessIntraLayerEvent (&hadEvent))
                return false;
        } while (hadEvent);

        //Check whether we are done
        myIntraRecv->communicationFinished(&isFinished);
    }
    while (!isFinished);

    return true;
}

/*EOF*/
